import tensorflow as tf
import numpy as np
import abc
import scipy.linalg
import scipy.sparse
import tensorflow_probability as tfp

from brainiak.matnormal.utils import (
    x_tx,
    xx_t,
    unflatten_cholesky_unique,
    flatten_cholesky_unique,
)
from brainiak.utils.kronecker_solvers import (
    tf_solve_lower_triangular_kron,
    tf_solve_upper_triangular_kron,
    tf_solve_lower_triangular_masked_kron,
    tf_solve_upper_triangular_masked_kron,
)

__all__ = [
    "CovBase",
    "CovIdentity",
    "CovAR1",
    "CovIsotropic",
    "CovDiagonal",
    "CovDiagonalGammaPrior",
    "CovUnconstrainedCholesky",
    "CovUnconstrainedCholeskyWishartReg",
    "CovUnconstrainedInvCholesky",
    "CovKroneckerFactored",
]


class CovBase(abc.ABC):
    """Base metaclass for residual covariances.
    For more on abstract classes, see
    https://docs.python.org/3/library/abc.html

    Parameters
    ----------

    size: int
        The size of the covariance matrix.

    """

    def __init__(self, size):
        self.size = size

        # Log-likelihood of this covariance (useful for regularization)
        self.logp = tf.constant(0, dtype=tf.float64)

    @abc.abstractmethod
    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized to fit
            this covariance
        """
        pass

    @property
    def logdet(self):
        """ log determinant of this covariance
        """
        pass

    @abc.abstractmethod
    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
            compute :math:`\\Sigma^{-1}x`
        """
        pass

    @property
    def _prec(self):
        """Expose the precision explicitly (mostly for testing /
        visualization, materializing large covariances may be intractable)
        """
        return self.solve(tf.eye(self.size, dtype=tf.float64))

    @property
    def _cov(self):
        """Expose the covariance explicitly (mostly for testing /
        visualization, materializing large covariances may be intractable)
        """
        return tf.linalg.inv(self._prec)


class CovIdentity(CovBase):
    """Identity noise covariance.
    """

    def __init__(self, size):
        super(CovIdentity, self).__init__(size)

    @property
    def logdet(self):
        return tf.constant(0.0, "float64")

    def get_optimize_vars(self):
        """Returns a list of tf variables that need to get optimized to
            fit this covariance
        """
        return []

    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
            compute :math:`\\Sigma^{-1}x`
        """
        return X

    @property
    def _prec(self):
        """Expose the precision explicitly (mostly for testing /
        visualization, materializing large covariances may be intractable)
        """
        return tf.eye(self.size, dtype=tf.float64)

    @property
    def _cov(self):
        """Expose the covariance explicitly (mostly for testing /
        visualization, materializing large covariances may be intractable)
        """
        return tf.eye(self.size, dtype=tf.float64)


class CovAR1(CovBase):
    """AR(1) covariance parameterized by autoregressive parameter rho
    and new noise sigma.

    Parameters
    ----------
    size: int
        size of covariance matrix
    rho: float or None
        initial value of autoregressive parameter (if None, initialize
        randomly)
    sigma: float or None
        initial value of new noise parameter (if None, initialize randomly)

    """

    def __init__(self, size, rho=None, sigma=None, scan_onsets=None):

        super(CovAR1, self).__init__(size)

        # Similar to BRSA trick I think
        if scan_onsets is None:
            self.run_sizes = [size]
            self.offdiag_template = tf.constant(
                scipy.linalg.toeplitz(np.r_[0, 1, np.zeros(size - 2)]),
                dtype=tf.float64
            )
            self.diag_template = tf.constant(
                np.diag(np.r_[0, np.ones(size - 2), 0]))
        else:
            self.run_sizes = np.ediff1d(np.r_[scan_onsets, size])
            sub_offdiags = [
                scipy.linalg.toeplitz(np.r_[0, 1, np.zeros(r - 2)])
                for r in self.run_sizes
            ]
            self.offdiag_template = tf.constant(
                scipy.sparse.block_diag(sub_offdiags).toarray()
            )
            subdiags = [np.diag(np.r_[0, np.ones(r - 2), 0])
                        for r in self.run_sizes]
            self.diag_template = tf.constant(
                scipy.sparse.block_diag(subdiags).toarray()
            )

        self._identity_mat = tf.constant(np.eye(size))

        if sigma is None:
            self.log_sigma = tf.Variable(
                tf.random.normal([1], dtype=tf.float64), name="log_sigma"
            )
        else:
            self.log_sigma = tf.Variable(np.log(sigma), name="log_sigma")

        if rho is None:
            self.rho_unc = tf.Variable(
                tf.random.normal([1], dtype=tf.float64), name="rho_unc"
            )
        else:
            self.rho_unc = tf.Variable(
                scipy.special.logit(rho / 2 + 0.5), name="rho_unc"
            )

    @property
    def logdet(self):
        """ log-determinant of this covariance
        """
        # first, unconstrain rho and sigma
        rho = 2 * tf.sigmoid(self.rho_unc) - 1
        # now compute logdet
        return tf.reduce_sum(
            input_tensor=2
            * tf.constant(self.run_sizes, dtype=tf.float64)
            * self.log_sigma
            - tf.math.log(1 - tf.square(rho))
        )

    @property
    def _prec(self):
        """Precision matrix corresponding to this AR(1) covariance.
        We assume stationarity within block so no special case
        for first/last element of a block. This makes constructing this
        matrix easier.
        reprsimil.BRSA says (I - rho1 * D + rho1**2 * F) / sigma**2 and we
        use the same trick
        """
        rho = 2 * tf.sigmoid(self.rho_unc) - 1
        sigma = tf.exp(self.log_sigma)

        return (
            self._identity_mat
            - rho * self.offdiag_template
            + rho ** 2 * self.diag_template
        ) / tf.square(sigma)

    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized to
        fit this covariance
        """
        return [self.rho_unc, self.log_sigma]

    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
        compute :math:`\\Sigma^{-1}x`
        """
        return tf.matmul(self._prec, X)


class CovIsotropic(CovBase):
    """Scaled identity (isotropic) noise covariance.

    Parameters
    ----------
    size: int
        size of covariance matrix
    var: float or None
        initial value of new variance parameter (if None, initialize randomly)

    """

    def __init__(self, size, var=None):
        super(CovIsotropic, self).__init__(size)
        if var is None:
            self.log_var = tf.Variable(
                tf.random.normal([1], dtype=tf.float64), name="sigma"
            )
        else:
            self.log_var = tf.Variable(np.log(var), name="sigma")
        self.var = tf.exp(self.log_var)

    @property
    def logdet(self):
        return self.size * self.log_var

    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized to fit
            this covariance
        """
        return [self.log_var]

    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
        compute :math:`\\Sigma^{-1}x`

        Parameters
        ----------
        X: tf.Tensor
            Tensor to multiply by inverse of this covariance

        """
        return X / self.var


class CovDiagonal(CovBase):
    """Uncorrelated (diagonal) noise covariance

    Parameters
    ----------
    size: int
        size of covariance matrix
    diag_var: float or None
        initial value of (diagonal) variance vector (if None, initialize
        randomly)

    """

    def __init__(self, size, diag_var=None):
        super(CovDiagonal, self).__init__(size)
        if diag_var is None:
            self.logprec = tf.Variable(
                tf.random.normal([size], dtype=tf.float64), name="precisions"
            )
        else:
            self.logprec = tf.Variable(
                np.log(1 / diag_var), name="log-precisions")

    @property
    def logdet(self):
        return -tf.reduce_sum(input_tensor=self.logprec)

    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized to fit
            this covariance
        """
        return [self.logprec]

    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
        compute :math:`\\Sigma^{-1}x`

        Parameters
        ----------
        X: tf.Tensor
            Tensor to multiply by inverse of this covariance

        """
        prec = tf.exp(self.logprec)
        prec_dimaugmented = tf.expand_dims(prec, -1)
        return tf.multiply(prec_dimaugmented, X)


class CovDiagonalGammaPrior(CovDiagonal):
    """Uncorrelated (diagonal) noise covariance
    """

    def __init__(self, size, sigma=None, alpha=1.5, beta=1e-10):
        super(CovDiagonalGammaPrior, self).__init__(size, sigma)

        self.ig = tfp.distributions.InverseGamma(
            concentration=tf.constant(alpha, dtype=tf.float64),
            scale=tf.constant(beta, dtype=tf.float64),
        )

        self.logp = tf.reduce_sum(
            input_tensor=self.ig.log_prob(tf.exp(self.logprec)))


class CovUnconstrainedCholesky(CovBase):
    """Unconstrained noise covariance parameterized in terms of its cholesky
    """

    def __init__(self, size=None, Sigma=None):

        if size is None and Sigma is None:
            raise RuntimeError("Must pass either Sigma or size but not both")

        if size is not None and Sigma is not None:
            raise RuntimeError("Must pass either Sigma or size but not both")

        if Sigma is not None:
            size = Sigma.shape[0]

        super(CovUnconstrainedCholesky, self).__init__(size)

        # number of parameters in the triangular mat
        npar = (size * (size + 1)) // 2

        if Sigma is None:
            self.L_flat = tf.Variable(
                tf.random.normal([npar], dtype=tf.float64), name="L_flat"
            )

        else:
            L = np.linalg.cholesky(Sigma)
            self.L_flat = tf.Variable(
                flatten_cholesky_unique(L), name="L_flat")

        self.optimize_vars = [self.L_flat]

    @property
    def L(self):
        """
        Cholesky factor of this covariance
        """
        return unflatten_cholesky_unique(self.L_flat)

    @property
    def logdet(self):
        return 2 * tf.reduce_sum(input_tensor=tf.math.log(
                                 tf.linalg.diag_part(self.L)))

    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized to fit
             this covariance
        """
        return [self.L_flat]

    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
        compute :math:`\\Sigma^{-1}x` (using cholesky solve)

        Parameters
        ----------
        X: tf.Tensor
            Tensor to multiply by inverse of this covariance

        """
        return tf.linalg.cholesky_solve(self.L, X)


class CovUnconstrainedCholeskyWishartReg(CovUnconstrainedCholesky):
    """Unconstrained noise covariance parameterized in terms of its
    cholesky factor. Regularized using the trick from
    Chung et al. 2015 such that as the covariance approaches
    singularity, the likelihood goes to 0.

    References
    ----------
    Chung, Y., Gelman, A., Rabe-Hesketh, S., Liu, J., & Dorie, V. (2015).
    Weakly Informative Prior for Point Estimation of Covariance Matrices
    in Hierarchical Models. Journal of Educational and Behavioral Statistics,
    40(2), 136–157. https://doi.org/10.3102/1076998615570945
    """

    def __init__(self, size, Sigma=None):
        super(CovUnconstrainedCholeskyWishartReg, self).__init__(size)
        self.wishartReg = tfp.distributions.WishartTriL(
            df=tf.constant(size + 2, dtype=tf.float64),
            scale_tril=tf.constant(1e5 * np.eye(size), dtype=tf.float64),
        )

        Sigma = xx_t(self.L)
        self.logp = self.wishartReg.log_prob(Sigma)


class CovUnconstrainedInvCholesky(CovBase):
    """Unconstrained noise covariance parameterized
       in terms of its precision cholesky. Use this over the
       regular cholesky unless you have a good reason not to, since
       this saves a cholesky solve on every step of optimization
    """

    def __init__(self, size=None, invSigma=None):

        if size is None and invSigma is None:
            raise RuntimeError(
                "Must pass either invSigma or size but not both")

        if size is not None and invSigma is not None:
            raise RuntimeError(
                "Must pass either invSigma or size but not both")

        if invSigma is not None:
            size = invSigma.shape[0]

        super(CovUnconstrainedInvCholesky, self).__init__(size)

        # number of parameters in the triangular mat
        npar = (size * (size + 1)) // 2

        if invSigma is None:
            self.Linv_flat = tf.Variable(
                tf.random.normal([npar], dtype=tf.float64), name="Linv_flat"
            )

        else:
            Linv = np.linalg.cholesky(invSigma)
            self.Linv_flat = tf.Variable(
                flatten_cholesky_unique(Linv), name="Linv_flat"
            )

    @property
    def Linv(self):
        """
        Inverse of Cholesky factor of this covariance
        """
        return unflatten_cholesky_unique(self.Linv_flat)

    @property
    def logdet(self):
        return -2 * tf.reduce_sum(
            input_tensor=tf.math.log(tf.linalg.diag_part(self.Linv))
        )

    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized to fit
            this covariance
        """
        return [self.Linv_flat]

    def solve(self, X):
        """Given this covariance :math:`\\Sigma` and some input :math:`X`,
        compute :math:`\\Sigma^{-1}x` (using cholesky solve)

        Parameters
        ----------
        X: tf.Tensor
            Tensor to multiply by inverse of this covariance

        """
        return tf.matmul(x_tx(self.Linv), X)


class CovKroneckerFactored(CovBase):
    """ Kronecker product noise covariance parameterized in terms
    of its component cholesky factors
    """

    def __init__(self, sizes, Sigmas=None, mask=None):
        """Initialize the kronecker factored covariance object.

        Arguments
        ---------
        sizes : list
            List of dimensions (int) of the factors
            E.g. ``sizes = [2, 3]`` will create two factors of
            sizes 2x2 and 3x3 giving us a 6x6 dimensional covariance
        Sigmas : list (default : None)
            Initial guess for the covariances. List of positive definite
            covariance matrices the same sizes as sizes.
        mask : int array (default : None)
            1-D tensor with length equal to product of sizes with 1 for
            valid elements and 0 for don't care

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If sizes is not a list
        """
        if not isinstance(sizes, list):
            raise TypeError("sizes is not a list")

        self.sizes = sizes
        self.nfactors = len(sizes)
        self.size = np.prod(np.array(sizes), dtype=np.int32)

        npar = [(size * (size + 1)) // 2 for size in self.sizes]
        if Sigmas is None:
            self.Lflat = [
                tf.Variable(
                    tf.random.normal([npar[i]], dtype=tf.float64),
                    name="L" + str(i) + "_flat",
                )
                for i in range(self.nfactors)
            ]
        else:
            self.Lflat = [
                tf.Variable(
                    flatten_cholesky_unique(np.linalg.cholesky(Sigmas[i])),
                    name="L" + str(i) + "_flat",
                )
                for i in range(self.nfactors)
            ]
        self.mask = mask

    @property
    def L(self):
        return [unflatten_cholesky_unique(mat) for mat in self.Lflat]

    def get_optimize_vars(self):
        """ Returns a list of tf variables that need to get optimized
            to fit this covariance
        """
        return self.Lflat

    @property
    def logdet(self):
        """ log|Sigma| using the diagonals of the cholesky factors.
        """
        if self.mask is None:
            n_list = tf.stack(
                [tf.cast(tf.shape(input=mat)[0], dtype=tf.float64)
                 for mat in self.L]
            )
            n_prod = tf.reduce_prod(input_tensor=n_list)
            logdet = tf.stack(
                [
                    tf.reduce_sum(
                        input_tensor=tf.math.log(
                            tf.linalg.tensor_diag_part(mat))
                    )
                    for mat in self.L
                ]
            )
            logdetfinal = tf.reduce_sum(
                input_tensor=(logdet * n_prod) / n_list)
        else:
            n_list = [tf.shape(input=mat)[0] for mat in self.L]
            mask_reshaped = tf.reshape(self.mask, n_list)
            logdet = 0.0
            for i in range(self.nfactors):
                indices = list(range(self.nfactors))
                indices.remove(i)
                logdet += (tf.math.log(tf.linalg.tensor_diag_part(self.L[i])) *
                           tf.cast(
                            tf.reduce_sum(
                                input_tensor=mask_reshaped, axis=indices),
                           dtype=tf.float64,
                           ))
            logdetfinal = tf.reduce_sum(input_tensor=logdet)
        return 2.0 * logdetfinal

    def solve(self, X):
        """ Given this covariance :math:`\\Sigma` and some input :math:`X`,
        compute :math:`\\Sigma^{-1}x` using traingular solves with the cholesky
        factors.

        Specifically, we solve :math:`L L^T x = y` by solving
        :math:`L z = y` and :math:`L^T x = z`.

        Parameters
        ----------
        X: tf.Tensor
            Tensor to multiply by inverse of this covariance

        """
        if self.mask is None:
            z = tf_solve_lower_triangular_kron(self.L, X)
            x = tf_solve_upper_triangular_kron(self.L, z)
        else:
            z = tf_solve_lower_triangular_masked_kron(self.L, X, self.mask)
            x = tf_solve_upper_triangular_masked_kron(self.L, z, self.mask)
        return x



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
import logging
import numpy as np
from brainiak.fcma.util import compute_correlation
from brainiak.fcma.preprocessing import generate_epochs_info
from brainiak import image, io
import scipy.io

format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
# if want to output log to a file instead of outputting log to the console,
# replace "stream=sys.stdout" with "filename='fcma.log'"
logging.basicConfig(level=logging.INFO, format=format, stream=sys.stdout)
logger = logging.getLogger(__name__)

# python3 corr_comp.py face_scene bet.nii.gz face_scene/prefrontal_top_mask.nii.gz face_scene/fs_epoch_labels.npy
if __name__ == '__main__':
    if len(sys.argv) != 5:
        logger.error('the number of input argument is not correct')
        sys.exit(1)

    data_dir = sys.argv[1]
    extension = sys.argv[2]
    mask_file = sys.argv[3]
    epoch_file = sys.argv[4]

    images = io.load_images_from_dir(data_dir, extension)
    mask = io.load_boolean_mask(mask_file)
    conditions = io.load_labels(epoch_file)
    raw_data = list(image.multimask_images(images, (mask,)))
    epoch_info = generate_epochs_info(conditions)

    for idx, epoch in enumerate(epoch_info):
        label = epoch[0]
        sid = epoch[1]
        start = epoch[2]
        end = epoch[3]
        mat = raw_data[sid][0][:, start:end]
        mat = np.ascontiguousarray(mat, dtype=np.float32)
        logger.info(
            'start to compute correlation for subject %d epoch %d with label %d' %
            (sid, idx, label)
        )
        corr = compute_correlation(mat, mat)
        mdict = {}
        mdict['corr'] = corr
        filename = str(label) + '_' + str(sid) + '_' + str(idx)
        logger.info(
            'start to write the correlation matrix to disk as %s' %
            filename
        )
        scipy.io.savemat(filename, mdict)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sklearn.exceptions import NotFittedError
import pytest


def test_can_instantiate():
    import brainiak.funcalign.rsrm
    s = brainiak.funcalign.rsrm.RSRM()
    assert s, "Invalid RSRM instance!"

    import numpy as np
    np.random.seed(0)

    voxels = 100
    samples = 500
    subjects = 2
    features = 3

    s = brainiak.funcalign.rsrm.RSRM(n_iter=5, features=features)
    assert s, "Invalid RSRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))

    X = []
    W = []
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    W.append(Q)
    X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        s.transform(X)
    print("Test: transforming before fitting the model")

    # Check that it does NOT run with 1 subject
    with pytest.raises(ValueError):
        s.fit(X)
    print("Test: running RSRM with 1 subject")

    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        W.append(Q)
        X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that runs with 2 subject
    s.fit(X)

    assert len(s.w_) == subjects, (
        "Invalid computation of RSRM! (wrong # subjects in W)")
    for subject in range(subjects):
        assert s.s_[subject].shape[0] == voxels, (
            "Invalid computation of RSRM! (wrong # voxels in S)")
        assert s.s_[subject].shape[1] == samples, (
            "Invalid computation of RSRM! (wrong # samples in S)")
        assert s.w_[subject].shape[0] == voxels, (
            "Invalid computation of RSRM! (wrong # voxels in W)")
        assert s.w_[subject].shape[1] == features, (
            "Invalid computation of RSRM! (wrong # features in W)")
        ortho = np.linalg.norm(s.w_[subject].T.dot(s.w_[subject])
                               - np.eye(s.w_[subject].shape[1]),
                               'fro')
        assert ortho < 1e-7, "A Wi mapping is not orthonormal in RSRM."
        difference = np.linalg.norm(X[subject] - s.w_[subject].dot(s.r_),
                                    'fro')
        datanorm = np.linalg.norm(X[subject], 'fro')
        assert difference/datanorm < 1.0, "Model seems incorrectly computed."
    assert s.r_.shape[0] == features, (
        "Invalid computation of RSRM! (wrong # features in R)")
    assert s.r_.shape[1] == samples, (
        "Invalid computation of RSRM! (wrong # samples in R)")

    # Check that it does run to compute the shared response after the model
    # computation
    new_r, _ = s.transform(X)

    assert len(new_r) == subjects, (
        "Invalid computation of RSRM! (wrong # subjects after transform)")
    for subject in range(subjects):
        assert new_r[subject].shape[0] == features, (
            "Invalid computation of RSRM! (wrong # features after transform)")
        assert new_r[subject].shape[1] == samples, (
            "Invalid computation of RSRM! (wrong # samples after transform)")

    # Check that it does run to compute a new subject
    new_w, new_s = s.transform_subject(X[0])
    assert new_w.shape[1] == features, (
            "Invalid computation of RSRM! (wrong # features for new subject)")
    assert new_s.shape[1] == samples, (
            "Invalid computation of RSRM! (wrong # samples for new subject)")
    assert new_s.shape[0] == voxels, (
            "Invalid computation of RSRM! (wrong # voxels for new subject)")
    assert new_w.shape[0] == voxels, (
            "Invalid computation of RSRM! (wrong # voxels for new subject)")

    # Check that it does NOT run with non-matching number of subjects
    with pytest.raises(ValueError):
        s.transform(X[1])
    print("Test: transforming with non-matching number of subjects")

    # Check that it does not run without enough samples (TRs).
    with pytest.raises(ValueError):
        s.set_params(features=(samples+1))
        s.fit(X)
    print("Test: not enough samples")

    # Check that it does not run with different number of samples (TRs)
    S2 = S[:, :-2]
    X.append(Q.dot(S2))
    with pytest.raises(ValueError):
        s.fit(X)
    print("Test: different number of samples per subject")



"""-----------------------------------------------------------------------------

finalize.py (Last Updated: 10/16/2020)

The purpose of this script is to finalize the rt-cloud session. Specifically,
here we want to dowload any important files from the cloud back to the stimulus
computer and maybe even delete files from the cloud that we don't want to 
stay there (maybe for privacy purposes).

-----------------------------------------------------------------------------"""

print(""
    "-----------------------------------------------------------------------------\n"
    "The finalize script has not been implemented for this project.\n\n"
    "To implement, please edit finalize.py. For an example as to how you can\n"
    "do this, refer to the finalize.py script in the sample project for the\n"
    "rtcloud framework. Go to https://github.com/brainiak/rt-cloud.git for the\n"
    "repo and navigate to projects/sample/\n"
    "-----------------------------------------------------------------------------")


"""Example of finding event segmentations on simulated data

This code generates simulated datasets that have temporally-clustered
structure (with the same series of latent event patterns). An event
segmentation is learned on the first dataset, and then we try to find the same
series of events in other datasets. We measure how well we find the latent
boundaries and the log-likelihood of the fits, and compare to a null model
in which the event order is randomly shuffled.
"""
import brainiak.eventseg.event
import numpy as np
from scipy import stats
import logging
import matplotlib.pyplot as plt

logging.basicConfig(level=logging.DEBUG)


def generate_event_labels(T, K, length_std):
    event_labels = np.zeros(T, dtype=int)
    start_TR = 0
    for e in range(K - 1):
        length = round(
            ((T - start_TR) / (K - e)) * (1 + length_std * np.random.randn()))
        length = min(max(length, 1), T - start_TR - (K - e))
        event_labels[start_TR:(start_TR + length)] = e
        start_TR = start_TR + length
    event_labels[start_TR:] = K - 1

    return event_labels


def generate_data(V, T, event_labels, event_means, noise_std):
    simul_data = np.empty((V, T))
    for t in range(T):
        simul_data[:, t] = stats.multivariate_normal.rvs(
            event_means[:, event_labels[t]], cov=noise_std, size=1)

    simul_data = stats.zscore(simul_data, axis=1, ddof=1)
    return simul_data


# Parameters for creating small simulated datasets
V = 10
K = 10
T = 500
T2 = 300

# Generate the first dataset
np.random.seed(1)
event_means = np.random.randn(V, K)
event_labels = generate_event_labels(T, K, 0.1)
simul_data = generate_data(V, T, event_labels, event_means, 1)

# Find the events in this dataset
simul_seg = brainiak.eventseg.event.EventSegment(K)
simul_seg.fit(simul_data.T)

# Generate other datasets with the same underlying sequence of event
# patterns, and try to find matching events
test_loops = 10
bound_match = np.empty((2, test_loops))
LL = np.empty((2, test_loops))
for test_i in range(test_loops):
    # Generate data
    event_labels2 = generate_event_labels(T2, K, 0.5)
    simul_data2 = generate_data(V, T2, event_labels2, event_means, 0.1)

    # Find events matching previously-learned events
    gamma, LL[0, test_i] = simul_seg.find_events(simul_data2.T)
    est_events2 = np.argmax(gamma, axis=1)
    bound_match[0, test_i] = 1 - np.sum(abs(np.diff(event_labels2) -
                                            np.diff(est_events2))) / (2 * K)

    # Run again, but with the order of events shuffled so that it no longer
    # corresponds to the training data
    gamma, LL[1, test_i] = simul_seg.find_events(simul_data2.T, scramble=True)
    est_events2 = np.argmax(gamma, axis=1)
    bound_match[1, test_i] = 1 - np.sum(abs(np.diff(event_labels2) -
                                            np.diff(est_events2))) / (2 * K)

# Across the testing datasets, print how well we identify the true event
# boundaries and the log-likehoods in real vs. shuffled data
print("Boundary match: {:.2} (null: {:.2})".format(
    np.mean(bound_match[0, :]), np.mean(bound_match[1, :])))
print("Log-likelihood: {:.3} (null: {:.3})".format(
    np.mean(LL[0, :]), np.mean(LL[1, :])))

plt.figure()
plt.subplot(2, 1, 1)
plt.imshow(simul_data2, interpolation='nearest', cmap=plt.cm.bone,
           aspect='auto')
plt.xlabel('Timepoints')
plt.ylabel('Voxels')
plt.subplot(2, 1, 2)
gamma, LL[0, test_i] = simul_seg.find_events(simul_data2.T)
est_events2 = np.argmax(gamma, axis=1)
plt.plot(est_events2)
plt.xlabel('Timepoints')
plt.ylabel('Event label')
plt.show()



"""-----------------------------------------------------------------------------

sample.py (Last Updated: 01/26/2021)

The purpose of this script is to run a sample project for the BrainIAK Aperture 
paper. This sample project script will be wrapped by the projectInterface and 
it will be made accessible on the Web Server.

The purpose of this *particular* script is to demonstrated how you can use the
various scripts, functions, etc. we have developed for your use! The functions
we will reference live in 'rt-cloud/rtCommon/'.

Finally, this script is called from the projectServer which is started in the
jupyter notebook.

-----------------------------------------------------------------------------"""


# add the full path to the rtcloud
import os
path_to_rtcloud = os.getenv('RTCLOUD_PATH')
if path_to_rtcloud == None:
    print("Please set RTCLOUD_PATH, see instructions")
    raise ValueError

# import other important modules
import sys
import time
import argparse
import numpy as np
import nibabel as nib
import scipy.io as sio
from nilearn.input_data import NiftiMasker
from sklearn import preprocessing
from sklearn import svm
import nilearn
from nilearn.masking import apply_mask
from scipy.stats import zscore
from sklearn.preprocessing import StandardScaler

import warnings # ignore warnings when importing dicomreaders below
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=UserWarning)
    from nibabel.nicom import dicomreaders

# obtain full path for current directory: '.../rt-cloud/projects/sample'
currPath = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path_to_rtcloud)
# import project modules from rt-cloud
from rtCommon.utils import loadConfigFile
from rtCommon.clientInterface import ClientInterface
from rtCommon.imageHandling import readRetryDicomFromDataInterface, getDicomFileName, convertDicomImgToNifti, convertDicomFileToNifti, readNifti, convertDicomFileToNifti

def doRuns(cfg, dataInterface, subjInterface, webInterface):
    """
    This function is called by 'main()' below. Here, we use the 'dataInterface'
    to read in dicoms (presumably from the scanner, but here it's from a folder
    with previously collected dicom files), doing some sort of analysis in the
    cloud, and then sending the info to the web browser.

    INPUT:
        [1] cfg (configuration file with important variables)
        [2] dataInterface (this will allow a script from the cloud to access files
               from the stimulus computer, which receives dicom files directly
               from the Siemens console computer)
        [3] subjInterface - this allows sending feedback (e.g. classification results)
                to a subjectService running on the presentation computer to provide
                feedback to the subject (and optionally get their response).
    OUTPUT:
        None.
    """

    # variables we'll use throughout
    scanNum = cfg.scanNum[0]
    runNum = cfg.runNum[0]

    # obtain the path for the directory where the subject's dicoms live
    cfg.dicomDir = cfg.imgDir
    print("Location of the subject's dicoms: %s\n" %cfg.dicomDir)

    # Initialize a watch for the entire dicom folder using the function 'initWatch'
    # in dataInterface. (Later we will use watchFile() to look for a specific dicom)
    #   INPUT:
    #       [1] cfg.dicomDir (where the subject's dicom files live)
    #       [2] cfg.dicomNamePattern (the naming pattern of dicom files)
    #       [3] cfg.minExpectedDicomSize (a check on size to make sure we don't
    #               accidentally grab a dicom before it's fully acquired)
    dataInterface.initWatch(cfg.dicomDir, cfg.dicomNamePattern,
        cfg.minExpectedDicomSize)

    #We will use the function plotDataPoint in webInterface whenever we
    #  want to send values to the web browser so that they can be plotted in the
    #  --Data Plots-- tab.
    #However at the start of a run we will want to clear the plot, and we can use
    #clearRunPlot(runId), or clearAllPlots() also in the webInterface object.
    print("\n\nClear any pre-existing plot for this run using 'clearRunPlot(runNum)'")
    webInterface.clearRunPlot(runNum)
    print(''
          '###################################################################################')

    # declare the total number of TRs
    num_trainingData = cfg.numTrainingTRs
    num_total_TRs = cfg.numSynthetic

    for this_TR in np.arange(num_total_TRs):
        # declare variables that are needed to use 'readRetryDicomFromDataInterface'
        timeout_file = 5 # small number because of demo, can increase for real-time

        # use 'getDicomFileName' from 'imageHandling.py' to obtain the filename 
        #   of the dicom data you want to get... which is useful considering how
        #   complicated these filenames can be!
        #   INPUT:
        #       [1] cfg (config parameters)
        #       [2] scanNum (scan number)
        #       [3] fileNum (TR number, which will reference the correct file)
        #   OUTPUT:
        #       [1] fullFileName (the filename of the dicom that should be grabbed)
        fileName = getDicomFileName(cfg, scanNum, this_TR)

        # Use 'readRetryDicomFromDataInterface' in 'imageHandling.py' to wait for dicom
        #    files to be written by the scanner (uses 'watchFile' internally) and then
        #    reading the dicom file once it is available.
        #INPUT:
        #    [1] dataInterface (allows a cloud script to access files from the
        #        control room computer)
        #    [2] filename (the dicom file we're watching for and want to load)
        #    [3] timeout (time spent waiting for a file before timing out)
        #OUTPUT:
        #    [1] dicomData (with class 'pydicom.dataset.FileDataset')
        dicomData = readRetryDicomFromDataInterface(dataInterface, fileName,
            timeout_file)

        # We can use the 'convertDicomFileToNifti' function with dicomData as 
        #   input to get the data in NifTi format
        base, ext = os.path.splitext(fileName)
        assert ext == '.dcm'
        niftiFilename = base + '.nii'
        convertDicomFileToNifti(fileName, niftiFilename)
        niftiObject = readNifti(niftiFilename)
        
        # declare various things if it's the first TR
        if this_TR == 0:
            # load the labels and mask
            labels = np.load(os.path.join(cfg.imgDir, 'labels.npy'))
            print(os.path.join(cfg.imgDir, 'labels.npy'))
            ROI_nib = nib.load(os.path.join(currPath,'ROI_mask.nii.gz'))
            ROI_mask = np.array(ROI_nib.dataobj)
            mask_nib = nib.Nifti1Image(ROI_mask.T, affine=niftiObject.affine)
            # declare the number of TRs we will shift the data to account for the hemodynamic lag
            num_shiftTRs = 3
            # shift the labels to account for hemodynamic lag
            shifted_labels = np.concatenate([np.full((num_shiftTRs, 1), np.nan),labels])
            # set up a matrix that will hold all of the preprocessed data
            preprocessed_data = np.full((num_total_TRs, int(ROI_mask.sum())), np.nan)

        # preprocess the training data by applying the mask
        preprocessed_data[this_TR,:] = np.ravel(apply_mask(niftiObject,mask_nib).reshape(int(ROI_mask.sum()),1))

        ## Now we divide into one of three possible steps

        # STEP 1: collect the training data
        if this_TR < num_trainingData:
            print('Collected training data for TR %s' %this_TR)
        # STEP 2: train the SVM model
        elif this_TR == num_trainingData:
            print(''
                  '###################################################################################')
            print('Done collecting training data! \nStart training the classifier.')

            # snapshot of time to keep track of how long it takes to train the classifier
            start_time = time.time()

            scaler = StandardScaler()
            X_train = preprocessed_data[num_shiftTRs:num_trainingData]
            y_train = shifted_labels[num_shiftTRs:num_trainingData].reshape(-1,)
            # we don't want to include rest data
            X_train_noRest = X_train[y_train != 0]
            y_train_noRest = y_train[y_train != 0]
            # and we want to zscore the bold data
            X_train_noRest_zscored = scaler.fit_transform(X_train_noRest)
            clf = svm.SVC(kernel='linear', C=0.01, class_weight='balanced')
            clf.fit(X_train_noRest_zscored, y_train_noRest)

            # print out the amount of time it took to train the classifier
            print('Classifier done training! Time it took: %.2f s' %(time.time() - start_time))
            print(''
                  '###################################################################################')
        elif this_TR > num_trainingData:
            # apply the classifier to new data to obtain prediction IF not a rest trial
            if shifted_labels[this_TR] != 0:
                prediction = clf.predict(scaler.transform(preprocessed_data[this_TR,:].reshape(1,-1)))
                print(f'Plotting classifier prediction for TR {this_TR}: {prediction}')
                webInterface.plotDataPoint(runNum, int(this_TR), float(prediction))

                # To send feedback for the subject at the presentation computer use
                #   the subjInterface as shown below. For this demo there is no
                #   presentation computer so we won't actually it.
                # subjInterface.setResult(runNum, int(this_TR), float(prediction))
            else:
                print(f'Skipping classification because it is a rest trial')


    X_test = preprocessed_data[num_trainingData+1:]
    y_test = shifted_labels[num_trainingData+1:num_total_TRs].reshape(-1,)
    # we don't want to include the rest data
    X_test_noRest = X_test[y_test != 0]
    y_test_noRest = y_test[y_test != 0]
    # and we want to zscore the bold data
    X_test_noRest_zscored = scaler.transform(X_test_noRest)
    accuracy_score = clf.score(X_test_noRest_zscored, y_test_noRest)
    print('Accuracy of classifier on new data: %s' %accuracy_score)
    # print(y_test_noRest)
    # print(clf.predict(X_test_noRest_zscored))
  
    print(""
    "###################################################################################\n"
    "REAL-TIME EXPERIMENT COMPLETE!")

    return

def main(argv=None):
    """
    This is the main function that is called when you run 'sample.py'.

    Here, you will load the configuration settings specified in the toml configuration
    file. It will initiate the class clientInterface which automatically connects
    to the projectServer and allows making requests, such as to get DICOM or NifTi
    data using the dataInterface contained in clientInterface.
    It will then call the function 'doRuns' to actually start doing the experiment.
    """

    # define the parameters that will be recognized later on to set up fileIterface
    argParser = argparse.ArgumentParser()
    argParser.add_argument('--config', '-c', default=None, type=str,
                           help='experiment config file (.json or .toml)')
    argParser.add_argument('--runs', '-r', default='', type=str,
                           help='Comma separated list of run numbers')
    argParser.add_argument('--scans', '-s', default='', type=str,
                           help='Comma separated list of scan number')
    args = argParser.parse_args(argv)

    # load the experiment configuration file
    cfg = loadConfigFile(args.config)

    # obtain paths for important directories (e.g. location of dicom files)
    if cfg.imgDir is None:
        cfg.imgDir = os.path.join(currPath, 'dicomDir')
    cfg.codeDir = currPath

    # Make an RPC connection to the projectServer
    # The 'dataInterface' class allows you to read and write files from the
    #   control room computer to this script running in the cloud.
    # The 'subjInterface' class will allow us to send classification results
    #   as feedback to the subject at the presentation computer.
    clientRPC = ClientInterface()
    dataInterface = clientRPC.dataInterface
    subjInterface = clientRPC.subjInterface
    webInterface = clientRPC.webInterface
    
    # now that we have the necessary variables, call the function 'doRuns' in order
    #   to actually start reading dicoms and doing your analyses of interest!
    #   INPUT:
    #       [1] cfg (configuration file with important variables)
    #       [2] dataInterface (this will allow a script from the cloud to access files
    #               from the stimulus computer that receives dicoms from the Siemens
    #               console computer)
    #       [3] subjInterface - this allows sending feedback (e.g. classification results)
    #            to a subjectService running on the presentation computer to provide
    #            feedback to the subject (and optionally get their response).
    #       [4] webInterface - this allows updating information on the experimenter webpage.
    #            For example to plot data points, or update status messages.
    doRuns(cfg, dataInterface, subjInterface, webInterface)

    return 0


if __name__ == "__main__":
    """
    If 'sample.py' is invoked as a program, then actually go through all of the portions
    of this script. This statement is not satisfied if functions are called from another
    script using "from sample.py import FUNCTION"
    """
    main()
    sys.exit(0)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Distributed Searchlight SRM Example.

This example runs searchlight srm on time segment matching experiment using Sherlock
dataset. 

Example Usage
-------
If run 4 ranks:
    $ mpirun -n 4 python3 searchlight_srm_example.py

Author
-------
Hejia Zhang (Princeton University ELE Department)

Notes
-------
It's an implementation of:
Zhang, Hejia, et al. "A Searchlight Factor Model Approach for Locating Shared 
Information in Multi-Subject fMRI Analysis." arXiv preprint arXiv:1609.09432 (2016).
https://arxiv.org/abs/1609.09432

"""

import numpy as np
from mpi4py import MPI
import sys
import scipy.io as sio
from scipy.stats import stats
from brainiak.searchlight.searchlight import Searchlight
from brainiak.funcalign.srm import SRM
from brainiak.fcma.util import compute_correlation
import warnings

# parameters
sl_rad = 1 #searchlight length (of each edge) will be 1+2*sl_rad
nfeature = 10 #number of features in SRM for each searchlight
niter = 10 #number of interations in SRM

# sanity check
if sl_rad <= 0:
    raise ValueError('sl_rad must be positive')
if nfeature > (1+2*sl_rad)**3:
    print ('nfeature truncated')
    nfeature = int((1+2*sl_rad)**3)

# MPI parameters, do not need to change
comm = MPI.COMM_WORLD
rank = comm.rank
size = comm.size

# load data
movie_file = sio.loadmat('data/sl_movie_data.mat')
movie_data = movie_file['data']

# Dataset size parameters
dim1,dim2,dim3,ntr,nsubj = movie_data.shape

# preprocess data, zscore and set NaN to 0
all_data = [] # first half train, second half test
for s in range(nsubj):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # To ignore "RuntimeWarning: invalid value encountered in true_divide"
        # There are some 0 voxels in the data which I have to keep, so there will be a warning 
        # when z-scoring the data. It should be safe to ignore this warning. If your data does
        # not contain 0 voxel, you can remove the 2 lines above
        train_tmp = np.nan_to_num(stats.zscore(movie_data[:,:,:,:int(ntr/2),s],axis=3,ddof=1))
        test_tmp = np.nan_to_num(stats.zscore(movie_data[:,:,:,int(ntr/2):,s],axis=3,ddof=1))
    all_data.append(np.concatenate((train_tmp,test_tmp),axis=3))

# print information
if rank == 0:
    print ('searchlight length is {}'.format(sl_rad))
    print ('number of features in SRM: {}'.format(nfeature))
    print ('number of subjects is: {}'.format(len(all_data)))
    print ('number of TR is: {}'.format(ntr))
    print ('brain data dimension is {}-by-{}-by-{}'.format(dim1,dim2,dim3))

# Generate mask: mask is a 3D binary array, with active voxels being 1. I simply set 
# all voxels to be active in this example, but you should set the mask to fit your ROI
# in practice.
mask = np.ones((dim1,dim2,dim3), dtype=bool)

# Create searchlight object
sl = Searchlight(sl_rad=sl_rad)

# Distribute data to processes
# the first argument of "distribute" is a list of 4D arrays, and each 4D array is data 
# from a single subject
sl.distribute(all_data, mask)
# broadcast something that should be shared by all ranks 
sl.broadcast([niter,nfeature])

# time segment matching experiment. Define your own experiment function here
def time_segment_matching_accuracy(data, win_size=6): 
    nsubjs = len(data)
    (ndim, nsample) = data[0].shape
    accu = np.zeros(shape=nsubjs)
    nseg = nsample - win_size 
    # mysseg prediction prediction
    trn_data = np.zeros((ndim*win_size, nseg),order='f')
    # the trn data also include the tst data, but will be subtracted when 
    # calculating A
    for m in range(nsubjs):
        for w in range(win_size):
            trn_data[w*ndim:(w+1)*ndim,:] += data[m][:,w:(w+nseg)]
    for tst_subj in range(nsubjs):
        tst_data = np.zeros((ndim*win_size, nseg),order='f')
        for w in range(win_size):
            tst_data[w*ndim:(w+1)*ndim,:] = data[tst_subj][:,w:(w+nseg)]

        A =  np.nan_to_num(stats.zscore((trn_data - tst_data),axis=0, ddof=1))
        B =  np.nan_to_num(stats.zscore(tst_data,axis=0, ddof=1))

        # compute correlation matrix
        corr_mtx = compute_correlation(B.T,A.T)

        for i in range(nseg):
            for j in range(nseg):
                if abs(i-j)<win_size and i != j :
                    corr_mtx[i,j] = -np.inf
        max_idx =  np.argmax(corr_mtx, axis=1)
        accu[tst_subj] = sum(max_idx == range(nseg)) / float(nseg)

    return accu

# Define voxel function: The function to be applied on each searchlight
def sfn(l, msk, myrad, bcast_var):
    # Arguments:
    # l -- a list of 4D arrays, containing data from a single searchlight
    # msk -- a 3D binary array, mask of this searchlight
    # myrad -- an integer, sl_rad
    # bcast_var -- whatever is broadcasted 

    # extract training and testing data
    train_data = []
    test_data = []
    d1,d2,d3,ntr = l[0].shape
    nvx = d1*d2*d3
    for s in l:
        train_data.append(np.reshape(s[:,:,:,:int(ntr/2)],(nvx,int(ntr/2))))
        test_data.append(np.reshape(s[:,:,:,int(ntr/2):],(nvx,ntr-int(ntr/2))))
    # train an srm model 
    srm = SRM(bcast_var[0],bcast_var[1])
    srm.fit(train_data)
    # transform test data
    shared_data = srm.transform(test_data)
    for s in range(len(l)):
        shared_data[s] = np.nan_to_num(stats.zscore(shared_data[s],axis=1,ddof=1))
    # run experiment
    accu = time_segment_matching_accuracy(shared_data)

    # return: can also return several values. In that case, the final output will be 
    # a 3D array of tuples
    return np.mean(accu) 

# Run searchlight
acc = sl.run_searchlight(sfn) # output is a 3D array in shape (dim1,dim2,dim3)

# save result
if rank == 0:
    print (acc)
    np.savez_compressed('data/searchlight_srm_tsm_acc.npz',acc=acc)  







import os
import glob
import pytest

from testbook import testbook


nb_files = glob.glob("docs/examples/**/*.ipynb", recursive=True)

mpi_notebooks = ["htfa", "FCMA", "SRM"]

nb_tests = []
for f in nb_files:
    # Mark notebooks that need MPI to skip for now,
    # we are having some issues on della
    if any([nb in f for nb in mpi_notebooks]):
        nb_tests.append(
            pytest.param(
                f,
                marks=pytest.mark.skip(
                    "notebooks that require MPI are WIP on della"
                ),
            )
        )
    elif "rtcloud" in f:
        nb_tests.append(
            pytest.param(
                f, marks=pytest.mark.skip("rtcloud is failing on della")
            )
        )
    elif "Matrix-normal" in f:
        nb_tests.append(
            pytest.param(
                f,
                marks=pytest.mark.skip(
                    "Matrix-normal notebook is flaky, disabled for now"
                ),
            )
        )
    else:
        nb_tests.append(f)  # type: ignore


# Helper function to mark specific notebooks as expected failure.
def mark_xfail(nb, **kwargs):
    nb_index = None
    for i, nb_file in enumerate(nb_files):
        if nb in nb_file:
            nb_index = i

    if nb_index is None:
        raise ValueError(
            f"Cannot set notebook {nb} to xfail because it could not be found"
        )

    nb_files[nb_index] = pytest.param(nb, marks=pytest.mark.xfail(**kwargs))


# mark_xfail('rtcloud_notebook.ipynb',
#           reason="Needs to have a web server installed, "
#                  "will probably need to run this in "
#                  "singularity on della")


@pytest.fixture(autouse=True)
def chdir_back_to_root():
    """
    This fixture sets up and tears down state before each example is run.
    Certain examples require that they are run from the local directory in
    which they reside. This changes directory. It reverses this after the
    test finishes.
    """

    # Get the current directory before running the test
    cwd = os.getcwd()

    yield

    # After the test, we need chdir back to root of the repo
    os.chdir(cwd)


@pytest.mark.notebook
@pytest.mark.parametrize("notebook_file", nb_tests)
def test_notebook(notebook_file):

    os.chdir(os.path.dirname(notebook_file))

    with testbook(os.path.basename(notebook_file), execute=True, timeout=3600):
        pass



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Hyper Parameter Optimization (HPO)

This implementation is based on the work in [Bergstra2011]_ and
[Bergstra2013]_.

.. [Bergstra2011] "Algorithms for Hyper-Parameter Optimization",
   James S. Bergstra and Bardenet, Rémi and Bengio, Yoshua
   and Kégl, Balázs. NIPS 2011

.. [Bergstra2013] "Making a Science of Model Search:
   Hyperparameter Optimization in Hundreds of Dimensions for
   Vision Architectures", James Bergstra, Daniel Yamins, David Cox.
   JMLR W&CP 28 (1) : 115–123, 2013

"""

# Authors: Narayanan Sundaram (Intel Labs)

import logging
import math
import numpy as np
from scipy.special import erf
import scipy.stats as st


logger = logging.getLogger(__name__)

__all__ = [
    "fmin",
]


def get_sigma(x, min_limit=-np.inf, max_limit=np.inf):
    """Compute the standard deviations around the points for a 1D GMM.

    We take the distance from the nearest left and right neighbors
    for each point, then use the max as the estimate of standard
    deviation for the gaussian mixture around that point.

    Arguments
    ---------
    x : 1D array
        Set of points to create the GMM

    min_limit : Optional[float], default : -inf
        Minimum limit for the distribution

    max_limit : Optional[float], default : inf
        maximum limit for the distribution

    Returns
    -------
    1D array
        Array of standard deviations
    """

    z = np.append(x, [min_limit, max_limit])
    sigma = np.ones(x.shape)
    for i in range(x.size):
        # Calculate the nearest left neighbor of x[i]
        # Find the minimum of (x[i] - k) for k < x[i]
        xleft = z[np.argmin([(x[i] - k) if k < x[i] else np.inf for k in z])]

        # Calculate the nearest right neighbor of x[i]
        # Find the minimum of (k - x[i]) for k > x[i]
        xright = z[np.argmin([(k - x[i]) if k > x[i] else np.inf for k in z])]

        sigma[i] = max(x[i] - xleft, xright - x[i])
        if sigma[i] == np.inf:
            sigma[i] = min(x[i] - xleft, xright - x[i])
        if (sigma[i] == -np.inf):  # should never happen
            sigma[i] = 1.0
    return sigma


class gmm_1d_distribution:
    """GMM 1D distribution.

    Given a set of points, we create this object so that we
    can calculate likelihoods and generate samples from this
    1D Gaussian mixture model.

    Attributes
    ----------
    points : 1D array
        Set of points to create the GMM

    N : int
        Number of points to create the GMM

    min_limit : Optional[float], default : -inf
        Minimum limit for the distribution

    max_limit : Optional[float], default : inf
        Maximum limit for the distribution

    weights : Optional[1D array], default : array of ones
        Used to weight the points non-uniformly if required
    """

    def __init__(self, x, min_limit=-np.inf, max_limit=np.inf, weights=1.0):
        self.points = x
        self.N = x.size
        self.min_limit = min_limit
        self.max_limit = max_limit
        self.sigma = get_sigma(x, min_limit=min_limit, max_limit=max_limit)
        self.weights = (2
                        / (erf((max_limit - x) / (np.sqrt(2.) * self.sigma))
                           - erf((min_limit - x) / (np.sqrt(2.) * self.sigma)))
                        * weights)
        self.W_sum = np.sum(self.weights)

    def get_gmm_pdf(self, x):
        """Calculate the GMM likelihood for a single point.

        .. math::
            y = \\sum_{i=1}^{N} w_i
            \\times \\text{normpdf}(x, x_i, \\sigma_i)/\\sum_{i=1}^{N} w_i
            :label: gmm-likelihood

        Arguments
        ---------
        x : float
            Point at which likelihood needs to be computed

        Returns
        -------
        float
            Likelihood value at x
        """

        def my_norm_pdf(xt, mu, sigma):
            z = (xt - mu) / sigma
            return (math.exp(-0.5 * z * z)
                    / (math.sqrt(2. * np.pi) * sigma))

        y = 0
        if (x < self.min_limit):
            return 0
        if (x > self.max_limit):
            return 0
        for _x in range(self.points.size):
            y += (my_norm_pdf(x, self.points[_x], self.sigma[_x])
                  * self.weights[_x]) / self.W_sum
        return y

    def __call__(self, x):
        """Return the GMM likelihood for given point(s).

        See :eq:`gmm-likelihood`.

        Arguments
        ---------
        x : scalar (or) 1D array of reals
            Point(s) at which likelihood needs to be computed

        Returns
        -------
        scalar (or) 1D array
            Likelihood values at the given point(s)
        """

        if np.isscalar(x):
            return self.get_gmm_pdf(x)
        else:
            return np.array([self.get_gmm_pdf(t) for t in x])

    def get_samples(self, n):
        """Sample the GMM distribution.

        Arguments
        ---------
        n : int
            Number of samples needed

        Returns
        -------
        1D array
            Samples from the distribution
        """

        normalized_w = self.weights / np.sum(self.weights)
        get_rand_index = st.rv_discrete(values=(range(self.N),
                                        normalized_w)).rvs(size=n)
        samples = np.zeros(n)
        k = 0
        j = 0
        while (k < n):
            i = get_rand_index[j]
            j = j + 1
            if (j == n):
                get_rand_index = st.rv_discrete(values=(range(self.N),
                                                normalized_w)).rvs(size=n)
                j = 0
            v = np.random.normal(loc=self.points[i], scale=self.sigma[i])
            if (v > self.max_limit or v < self.min_limit):
                continue
            else:
                samples[k] = v
                k = k + 1
                if (k == n):
                    break
        return samples


def get_next_sample(x, y, min_limit=-np.inf, max_limit=np.inf):
    """Get the next point to try, given the previous samples.

    We use [Bergstra2013]_ to compute the point that gives the largest
    Expected improvement (EI) in the optimization function. This model fits 2
    different GMMs - one for points that have loss values in the bottom 15%
    and another for the rest. Then we sample from the former distribution
    and estimate EI as the ratio of the likelihoods of the 2 distributions.
    We pick the point with the best EI among the samples that is also not
    very close to a point we have sampled earlier.

    Arguments
    ---------
    x : 1D array
        Samples generated from the distribution so far

    y : 1D array
        Loss values at the corresponding samples

    min_limit : float, default : -inf
        Minimum limit for the distribution

    max_limit : float, default : +inf
        Maximum limit for the distribution

    Returns
    -------
    float
        Next value to use for HPO
    """

    z = np.array(list(zip(x, y)), dtype=np.dtype([('x', float), ('y', float)]))
    z = np.sort(z, order='y')
    n = y.shape[0]
    g = int(np.round(np.ceil(0.15 * n)))
    ldata = z[0:g]
    gdata = z[g:n]
    lymin = ldata['y'].min()
    lymax = ldata['y'].max()
    weights = (lymax - ldata['y']) / (lymax - lymin)
    lx = gmm_1d_distribution(ldata['x'], min_limit=min_limit,
                             max_limit=max_limit, weights=weights)
    gx = gmm_1d_distribution(gdata['x'], min_limit=min_limit,
                             max_limit=max_limit)

    samples = lx.get_samples(n=1000)
    ei = lx(samples) / gx(samples)

    h = (x.max() - x.min()) / (10 * x.size)
    # assumes prior of x is uniform; should ideally change for other priors
    # d = np.abs(x - samples[ei.argmax()]).min()
    # CDF(x+d/2) - CDF(x-d/2) < 1/(10*x.size) then reject else accept
    s = 0
    while (np.abs(x - samples[ei.argmax()]).min() < h):
        ei[ei.argmax()] = 0
        s = s + 1
        if (s == samples.size):
            break
    xnext = samples[ei.argmax()]

    return xnext


def fmin(loss_fn,
         space,
         max_evals,
         trials,
         init_random_evals=30,
         explore_prob=0.2):
    """Find the minimum of function through hyper parameter optimization.

    Arguments
    ---------
    loss_fn : ``function(*args) -> float``
        Function that takes in a dictionary and returns a real value.
        This is the function to be minimized.

    space : dictionary
        Custom dictionary specifying the range and distribution of
        the hyperparamters.
        E.g. ``space = {'x': {'dist':scipy.stats.uniform(0,1),
        'lo':0, 'hi':1}}``
        for a 1-dimensional space with variable x in range [0,1]

    max_evals : int
        Maximum number of evaluations of loss_fn allowed

    trials : list
        Holds the output of the optimization trials.
        Need not be empty to begin with, new trials are appended
        at the end.

    init_random_evals : Optional[int], default 30
        Number of random trials to initialize the
        optimization.

    explore_prob : Optional[float], default 0.2
        Controls the exploration-vs-exploitation ratio. Value should
        be in [0,1]. By default, 20% of trails are random samples.

    Returns
    -------
    trial entry (dictionary of hyperparameters)
           Best hyperparameter setting found.
           E.g. {'x': 5.6, 'loss' : 0.5} where x is the best hyparameter
           value found and loss is the value of the function for the
           best hyperparameter value(s).

    Raises
    ------
    ValueError
        If the distribution specified in space does not support a ``rvs()``
        method to generate random numbers, a ValueError is raised.
    """

    for s in space:
        if not hasattr(space[s]['dist'], 'rvs'):
            raise ValueError('Unknown distribution type for variable')
        if 'lo' not in space[s]:
            space[s]['lo'] = -np.inf
        if 'hi' not in space[s]:
            space[s]['hi'] = np.inf

    if len(trials) > init_random_evals:
        init_random_evals = 0

    for t in range(max_evals):
        sdict = {}

        if t >= init_random_evals and np.random.random() > explore_prob:
            use_random_sampling = False
        else:
            use_random_sampling = True

        yarray = np.array([tr['loss'] for tr in trials])
        for s in space:
            sarray = np.array([tr[s] for tr in trials])
            if use_random_sampling:
                sdict[s] = space[s]['dist'].rvs()
            else:
                sdict[s] = get_next_sample(sarray, yarray,
                                           min_limit=space[s]['lo'],
                                           max_limit=space[s]['hi'])

        logger.debug('Explore' if use_random_sampling else 'Exploit')
        logger.info('Next point ', t, ' = ', sdict)

        y = loss_fn(sdict)
        sdict['loss'] = y
        trials.append(sdict)

    yarray = np.array([tr['loss'] for tr in trials])
    yargmin = yarray.argmin()

    logger.info('Best point so far = ', trials[yargmin])
    return trials[yargmin]



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Semi-Supervised Shared Response Model (SS-SRM)

The implementations are based on the following publications:

.. [Turek2016] "A Semi-Supervised Method for Multi-Subject fMRI Functional
   Alignment",
   J. S. Turek, T. L. Willke, P.-H. Chen, P. J. Ramadge
   IEEE International Conference on Acoustics, Speech and Signal Processing
   (ICASSP), 2017, pp. 1098-1102.
   https://doi.org/10.1109/ICASSP.2017.7952326
"""

# Authors: Javier Turek (Intel Labs), 2016

import logging

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin
from sklearn.utils import assert_all_finite
from sklearn.utils.validation import NotFittedError
from sklearn.utils.multiclass import unique_labels

import tensorflow as tf
from pymanopt.manifolds import Euclidean
from pymanopt.manifolds import Product
from pymanopt.optimizers import ConjugateGradient
from pymanopt import Problem, function
from pymanopt.manifolds import Stiefel

import gc

from brainiak.utils import utils
from brainiak.funcalign import srm

__all__ = [
    "SSSRM",
]

logger = logging.getLogger(__name__)


class SSSRM(BaseEstimator, ClassifierMixin, TransformerMixin):
    """Semi-Supervised Shared Response Model (SS-SRM)

    Given multi-subject data, factorize it as a shared response S among all
    subjects and an orthogonal transform W per subject, using also labeled
    data to train a Multinomial Logistic Regression (MLR) classifier (with
    l2 regularization) in a semi-supervised manner:

    .. math::
        (1-\\alpha) Loss_{SRM}(W_i,S;X_i)
        + \\alpha/\\gamma Loss_{MLR}(\\theta, bias; {(W_i^T \\times Z_i, y_i})
        + R(\\theta)
        :label: sssrm-eq

    (see Equations (1) and (4) in [Turek2016]_).

    Parameters
    ----------

    n_iter : int, default: 10
        Number of iterations to run the algorithm.

    features : int, default: 50
        Number of features to compute.

    gamma : float, default: 1.0
        Regularization parameter for the classifier.

    alpha : float, default: 0.5
        Balance parameter between the SRM term and the MLR term.

    rand_seed : int, default: 0
        Seed for initializing the random number generator.


    Attributes
    ----------

    w_ : list of array, element i has shape=[voxels_i, features]
        The orthogonal transforms (mappings) for each subject.

    s_ : array, shape=[features, samples]
        The shared response.

    theta_ : array, shape=[classes, features]
        The MLR class plane parameters.

    bias_ : array, shape=[classes]
        The MLR class biases.

    classes_ : array of int, shape=[classes]
        Mapping table for each classes to original class label.

    random_state_: `RandomState`
        Random number generator initialized using rand_seed

    Note
    ----

        The number of voxels may be different between subjects. However, the
        number of samples for the alignment data must be the same across
        subjects. The number of labeled samples per subject can be different.

        The Semi-Supervised Shared Response Model is approximated using the
        Block-Coordinate Descent (BCD) algorithm proposed in [Turek2016]_.

        This is a single node version.
    """

    def __init__(self, n_iter=10, features=50, gamma=1.0, alpha=0.5,
                 rand_seed=0):
        self.n_iter = n_iter
        self.features = features
        self.gamma = gamma
        self.alpha = alpha
        self.rand_seed = rand_seed
        return

    def fit(self, X, y, Z):
        """Compute the Semi-Supervised Shared Response Model

        Parameters
        ----------

        X : list of 2D arrays, element i has shape=[voxels_i, n_align]
            Each element in the list contains the fMRI data for alignment of
            one subject. There are n_align samples for each subject.

        y : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the data samples
            in Z.

        Z : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject
            for training the MLR classifier.

        """
        logger.info('Starting SS-SRM')

        # Check that the alpha value is in range (0.0,1.0)
        if 0.0 >= self.alpha or self.alpha >= 1.0:
            raise ValueError("Alpha parameter should be in range (0.0, 1.0)")

        # Check that the regularizer value is positive
        if 0.0 >= self.gamma:
            raise ValueError("Gamma parameter should be positive.")

        # Check the number of subjects
        if len(X) <= 1 or len(y) <= 1 or len(Z) <= 1:
            raise ValueError("There are not enough subjects in the input "
                             "data to train the model.")

        if not (len(X) == len(y)) or not (len(X) == len(Z)):
            raise ValueError("Different number of subjects in data.")

        # Check for input data sizes
        if X[0].shape[1] < self.features:
            raise ValueError(
                "There are not enough samples to train the model with "
                "{0:d} features.".format(self.features))

        # Check if all subjects have same number of TRs for alignment
        # and if alignment and classification data have the same number of
        # voxels per subject. Also check that there labels for all the classif.
        # sample
        number_trs = X[0].shape[1]
        number_subjects = len(X)
        for subject in range(number_subjects):
            assert_all_finite(X[subject])
            assert_all_finite(Z[subject])
            if X[subject].shape[1] != number_trs:
                raise ValueError("Different number of alignment samples "
                                 "between subjects.")
            if X[subject].shape[0] != Z[subject].shape[0]:
                raise ValueError("Different number of voxels between alignment"
                                 " and classification data (subject {0:d})"
                                 ".".format(subject))
            if Z[subject].shape[1] != y[subject].size:
                raise ValueError("Different number of samples and labels in "
                                 "subject {0:d}.".format(subject))

        # Map the classes to [0..C-1]
        new_y = self._init_classes(y)

        # Run SS-SRM
        self.w_, self.s_, self.theta_, self.bias_ = self._sssrm(X, Z, new_y)

        return self

    def _init_classes(self, y):
        """Map all possible classes to the range [0,..,C-1]

        Parameters
        ----------

        y : list of arrays of int, each element has shape=[samples_i,]
            Labels of the samples for each subject


        Returns
        -------
        new_y : list of arrays of int, each element has shape=[samples_i,]
            Mapped labels of the samples for each subject

        Note
        ----
            The mapping of the classes is saved in the attribute classes_.
        """
        self.classes_ = unique_labels(utils.concatenate_not_none(y))
        new_y = [None] * len(y)
        for s in range(len(y)):
            new_y[s] = np.digitize(y[s], self.classes_) - 1
        return new_y

    def transform(self, X, y=None):
        """Use the model to transform matrix to Shared Response space

        Parameters
        ----------

        X : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject
            note that number of voxels and samples can vary across subjects.

        y : not used as it only applies the mappings


        Returns
        -------

        s : list of 2D arrays, element i has shape=[features_i, samples_i]
            Shared responses from input data (X)
        """

        # Check if the model exist
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        # Check the number of subjects
        if len(X) != len(self.w_):
            raise ValueError("The number of subjects does not match the one"
                             " in the model.")

        s = [None] * len(X)
        for subject in range(len(X)):
            s[subject] = self.w_[subject].T.dot(X[subject])

        return s

    def predict(self, X):
        """Classify the output for given data

        Parameters
        ----------
        X : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject
            The number of voxels should be according to each subject at
            the moment of training the model.

        Returns
        -------
        p: list of arrays, element i has shape=[samples_i]
            Predictions for each data sample.
        """
        # Check if the model exist
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        # Check the number of subjects
        if len(X) != len(self.w_):
            raise ValueError("The number of subjects does not match the one"
                             " in the model.")

        X_shared = self.transform(X)
        p = [None] * len(X_shared)
        for subject in range(len(X_shared)):
            sumexp, _, exponents = utils.sumexp_stable(
                self.theta_.T.dot(X_shared[subject]) +
                self.bias_[:, np.newaxis])
            p[subject] = self.classes_[
                (exponents / sumexp[np.newaxis, :]).argmax(axis=0)]

        return p

    def _sssrm(self, data_align, data_sup, labels):
        """Block-Coordinate Descent algorithm for fitting SS-SRM.

        Parameters
        ----------

        data_align : list of 2D arrays, element i has shape=[voxels_i, n_align]
            Each element in the list contains the fMRI data for alignment of
            one subject. There are n_align samples for each subject.

        data_sup : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject for
            the classification task.

        labels : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the data samples
            in data_sup.


        Returns
        -------

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        s : array, shape=[features, samples]
            The shared response.
        """
        classes = self.classes_.size

        # Initialization:
        self.random_state_ = np.random.RandomState(self.rand_seed)
        random_states = [
            np.random.RandomState(self.random_state_.randint(2**32,
                                                             dtype=np.int64))
            for i in range(len(data_align))]
        # Set Wi's to a random orthogonal voxels by TRs
        w, _ = srm._init_w_transforms(data_align, self.features, random_states)

        # Initialize the shared response S
        s = SSSRM._compute_shared_response(data_align, w)

        # Initialize theta and bias
        theta, bias = self._update_classifier(data_sup, labels, w, classes)

        # calculate and print the objective function
        if logger.isEnabledFor(logging.INFO):
            objective = self._objective_function(data_align, data_sup, labels,
                                                 w, s, theta, bias)
            logger.info('Objective function %f' % objective)

        # Main loop:
        for iteration in range(self.n_iter):
            logger.info('Iteration %d' % (iteration + 1))

            # Update the mappings Wi
            w = self._update_w(data_align, data_sup, labels, w, s, theta, bias)

            # Output the objective function
            if logger.isEnabledFor(logging.INFO):
                objective = self._objective_function(data_align, data_sup,
                                                     labels, w, s, theta, bias)
                logger.info('Objective function after updating Wi  %f'
                            % objective)

            # Update the shared response S
            s = SSSRM._compute_shared_response(data_align, w)

            # Output the objective function
            if logger.isEnabledFor(logging.INFO):
                objective = self._objective_function(data_align, data_sup,
                                                     labels, w, s, theta, bias)
                logger.info('Objective function after updating S   %f'
                            % objective)

            # Update the MLR classifier, theta and bias
            theta, bias = self._update_classifier(data_sup, labels, w, classes)

            # Output the objective function
            if logger.isEnabledFor(logging.INFO):
                objective = self._objective_function(data_align, data_sup,
                                                     labels, w, s, theta, bias)
                logger.info('Objective function after updating MLR %f'
                            % objective)

        return w, s, theta, bias

    def _update_classifier(self, data, labels, w, classes):
        """Update the classifier parameters theta and bias

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject for
            the classification task.

        labels : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the data samples
            in data_sup.

        w : list of 2D array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        classes : int
            The number of classes in the classifier.


        Returns
        -------

        theta : array, shape=[features, classes]
            The MLR parameter for the class planes.

        bias : array shape=[classes,]
            The MLR parameter for class biases.
        """

        # Stack the data and labels for training the classifier
        data_stacked, labels_stacked, weights = \
            SSSRM._stack_list(data, labels, w)

        features = w[0].shape[1]
        total_samples = weights.size

        data_tf = tf.constant(data_stacked, dtype=tf.float64)
        labels_tf = tf.constant(labels_stacked, dtype=tf.int32)
        weights_tf = tf.constant(weights, dtype=tf.float64)
        constf2 = tf.constant(self.alpha / self.gamma, dtype=tf.float64)

        manifold = Product([Euclidean(features, classes),
                            Euclidean(1, classes)])

        @function.tensorflow(manifold)
        def cost(theta_tf, bias_tf):
            logits = tf.matmul(data_tf, theta_tf) + bias_tf
            log_p_y_given_x = tf.nn.log_softmax(logits)
            indices = tf.range(total_samples)
            gather_indices = tf.stack([indices, labels_tf], axis=1)
            log_probs = tf.gather_nd(log_p_y_given_x, gather_indices)
            f = (-constf2 * tf.reduce_sum(log_probs / weights_tf) +
                 0.5 * tf.reduce_sum(theta_tf ** 2))
            return f

        problem = Problem(manifold=manifold, cost=cost)
        solver = ConjugateGradient(min_gradient_norm=1e-6, verbosity=0)

        # Initialize variables
        theta_init = np.zeros((features, classes), dtype=np.float64)
        bias_init = np.zeros((1, classes), dtype=np.float64)
        solution = solver.run(problem, initial_point=[theta_init, bias_init])

        theta = solution.point[0]
        bias = solution.point[1].flatten()

        return theta, bias

    def _update_w(self, data_align, data_sup, labels, w, s, theta, bias):
        """

        Parameters
        ----------
        data_align : list of 2D arrays, element i has shape=[voxels_i, n_align]
            Each element in the list contains the fMRI data for alignment of
            one subject. There are n_align samples for each subject.

        data_sup : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject for
            the classification task.

        labels : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the data samples
            in data_sup.

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        s : array, shape=[features, samples]
            The shared response.

        theta : array, shape=[classes, features]
            The MLR class plane parameters.

        bias : array, shape=[classes]
            The MLR class biases.

        Returns
        -------

        w : list of 2D array, element i has shape=[voxels_i, features]
            The updated orthogonal transforms (mappings).
        """
        subjects = len(data_align)

        s_tf = tf.constant(s.astype(np.float64))
        theta_tf = tf.constant(theta.T.astype(np.float64))
        bias_tf = tf.constant(bias[np.newaxis, :].astype(np.float64))

        for subject in range(subjects):
            logger.info('Subject Wi %d' % subject)
            # Solve for subject i

            data_srm_subject_tf = tf.constant(
                data_align[subject].astype(np.float64))
            constf1 = tf.constant(
                (1 - self.alpha) * 0.5 / data_align[subject].shape[1],
                dtype=tf.float64)

            manifold = Stiefel(w[subject].shape[0], w[subject].shape[1])

            if data_sup[subject] is not None:
                data_sup_subject_tf = tf.constant(
                    data_sup[subject].astype(np.float64))
                labels_tf_subj = tf.constant(labels[subject], dtype=tf.int32)
                constf2 = tf.constant(
                    -self.alpha / self.gamma / data_sup[subject].shape[1],
                    dtype=tf.float64)

                @function.tensorflow(manifold)
                def cost(w_tf):
                    diff = (data_srm_subject_tf - tf.matmul(w_tf, s_tf))
                    f1 = constf1 * tf.reduce_sum(diff ** 2)
                    w_by_data = tf.matmul(tf.transpose(w_tf),
                                          data_sup_subject_tf)
                    logits = tf.transpose(
                        tf.matmul(theta_tf, w_by_data)) + bias_tf
                    log_p_y_given_x = tf.nn.log_softmax(logits)
                    indices = tf.range(data_sup_subject_tf.shape[1],
                                       dtype=tf.int32)
                    gather_indices = tf.stack([indices, labels_tf_subj],
                                              axis=1)
                    log_probs = tf.gather_nd(log_p_y_given_x, gather_indices)
                    f2 = constf2 * tf.reduce_sum(log_probs)
                    return f1 + f2
            else:
                @function.tensorflow(manifold)
                def cost(w_tf):
                    diff = (data_srm_subject_tf - tf.matmul(w_tf, s_tf))
                    f1 = constf1 * tf.reduce_sum(diff ** 2)
                    return f1

            f_subject = self._objective_function_subject(data_align[subject],
                                                         data_sup[subject],
                                                         labels[subject],
                                                         w[subject],
                                                         s, theta, bias)
            minstep = np.amin(((10 ** -np.floor(np.log10(f_subject))), 1e-1))

            problem = Problem(manifold=manifold, cost=cost)
            solver = ConjugateGradient(min_gradient_norm=1e-2,
                                       min_step_size=minstep,
                                       verbosity=0)
            w_init = w[subject].astype(np.float64)
            result = solver.run(problem, initial_point=w_init)
            w[subject] = result.point

        # Run garbage collector to avoid filling up the memory
        gc.collect()
        return w

    @staticmethod
    def _compute_shared_response(data, w):
        """ Compute the shared response S

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        w : list of 2D arrays, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.


        Returns
        -------

        s : array, shape=[features, samples]
            The shared response for the subjects data with the mappings in w.
        """
        s = np.zeros((w[0].shape[1], data[0].shape[1]))
        for m in range(len(w)):
            s = s + w[m].T.dot(data[m])
        s /= len(w)
        return s

    def _objective_function(self, data_align, data_sup, labels, w, s, theta,
                            bias):
        """Compute the objective function of the Semi-Supervised SRM

        See :eq:`sssrm-eq`.

        Parameters
        ----------

        data_align : list of 2D arrays, element i has shape=[voxels_i, n_align]
            Each element in the list contains the fMRI data for alignment of
            one subject. There are n_align samples for each subject.

        data_sup : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject for
            the classification task.

        labels : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the data samples
            in data_sup.

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        s : array, shape=[features, samples]
            The shared response.

        theta : array, shape=[classes, features]
            The MLR class plane parameters.

        bias : array, shape=[classes]
            The MLR class biases.


        Returns
        -------

        f_val : float
            The SS-SRM objective function evaluated based on the parameters to
            this function.
        """
        subjects = len(data_align)

        # Compute the SRM  loss
        f_val = 0.0
        for subject in range(subjects):
            samples = data_align[subject].shape[1]
            f_val += (1 - self.alpha) * (0.5 / samples) \
                * np.linalg.norm(data_align[subject] - w[subject].dot(s),
                                 'fro')**2

        # Compute the MLR loss
        f_val += self._loss_lr(data_sup, labels, w, theta, bias)

        return f_val

    def _objective_function_subject(self, data_align, data_sup, labels, w, s,
                                    theta, bias):
        """Compute the objective function for one subject.

        .. math:: (1-C)*Loss_{SRM}_i(W_i,S;X_i)
        .. math:: + C/\\gamma * Loss_{MLR_i}(\\theta, bias; {(W_i^T*Z_i, y_i})
        .. math:: + R(\\theta)

        Parameters
        ----------

        data_align : 2D array, shape=[voxels_i, samples_align]
            Contains the fMRI data for alignment of subject i.

        data_sup : 2D array, shape=[voxels_i, samples_i]
            Contains the fMRI data of one subject for the classification task.

        labels : array of int, shape=[samples_i]
            The labels for the data samples in data_sup.

        w : array, shape=[voxels_i, features]
            The orthogonal transform (mapping) :math:`W_i` for subject i.

        s : array, shape=[features, samples]
            The shared response.

        theta : array, shape=[classes, features]
            The MLR class plane parameters.

        bias : array, shape=[classes]
            The MLR class biases.


        Returns
        -------

        f_val : float
            The SS-SRM objective function for subject i evaluated on the
            parameters to this function.
        """
        # Compute the SRM  loss
        f_val = 0.0
        samples = data_align.shape[1]
        f_val += (1 - self.alpha) * (0.5 / samples) \
            * np.linalg.norm(data_align - w.dot(s), 'fro')**2

        # Compute the MLR loss
        f_val += self._loss_lr_subject(data_sup, labels, w, theta, bias)

        return f_val

    def _loss_lr_subject(self, data, labels, w, theta, bias):
        """Compute the Loss MLR for a single subject (without regularization)

        Parameters
        ----------

        data : array, shape=[voxels, samples]
            The fMRI data of subject i for the classification task.

        labels : array of int, shape=[samples]
            The labels for the data samples in data.

        w : array, shape=[voxels, features]
            The orthogonal transform (mapping) :math:`W_i` for subject i.

        theta : array, shape=[classes, features]
            The MLR class plane parameters.

        bias : array, shape=[classes]
            The MLR class biases.

        Returns
        -------

        loss : float
            The loss MLR for the subject
        """
        if data is None:
            return 0.0

        samples = data.shape[1]

        thetaT_wi_zi_plus_bias = (theta.T.dot(w.T.dot(data)) +
                                  bias[:, np.newaxis])
        sum_exp, max_value, _ = utils.sumexp_stable(thetaT_wi_zi_plus_bias)
        sum_exp_values = np.log(sum_exp) + max_value

        aux = 0.0
        for sample in range(samples):
            label = labels[sample]
            aux += thetaT_wi_zi_plus_bias[label, sample]
        return self.alpha / samples / self.gamma * (sum_exp_values.sum() - aux)

    def _loss_lr(self, data, labels, w, theta, bias):
        """Compute the Loss MLR (with the regularization)

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject for
            the classification task.

        labels : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the samples in
            data.

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        theta : array, shape=[classes, features]
            The MLR class plane parameters.

        bias : array, shape=[classes]
            The MLR class biases.


        Returns
        -------

        loss : float
            The loss MLR for the SS-SRM model
        """
        subjects = len(data)
        loss = 0.0
        for subject in range(subjects):
            if labels[subject] is not None:
                loss += self._loss_lr_subject(data[subject], labels[subject],
                                              w[subject], theta, bias)

        return loss + 0.5 * np.linalg.norm(theta, 'fro')**2

    @staticmethod
    def _stack_list(data, data_labels, w):
        """Construct a numpy array by stacking arrays in a list

        Parameter
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject for
            the classification task.

        data_labels : list of arrays of int, element i has shape=[samples_i]
            Each element in the list contains the labels for the samples in
            data.

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.


        Returns
        -------

        data_stacked : 2D array, shape=[samples, features]
            The data samples from all subjects are stacked into a single
            2D array, where "samples" is the sum of samples_i.

        labels_stacked : array, shape=[samples,]
            The labels from all subjects are stacked into a single
            array, where "samples" is the sum of samples_i.

        weights : array, shape=[samples,]
            The number of samples of the subject that are related to that
            sample. They become a weight per sample in the MLR loss.
        """
        labels_stacked = utils.concatenate_not_none(data_labels)

        weights = np.empty((labels_stacked.size,))
        data_shared = [None] * len(data)
        curr_samples = 0
        for s in range(len(data)):
            if data[s] is not None:
                subject_samples = data[s].shape[1]
                curr_samples_end = curr_samples + subject_samples
                weights[curr_samples:curr_samples_end] = subject_samples
                data_shared[s] = w[s].T.dot(data[s])
                curr_samples += data[s].shape[1]

        data_stacked = utils.concatenate_not_none(data_shared, axis=1).T
        return data_stacked, labels_stacked, weights
